import math
from itertools import groupby
from math import pi, cos

EARTH = 6378.137  # Radius of the earth in kilometer
METER = (1 / ((2 * pi / 360) * EARTH)) / 1000  # 1 meter in degree

# find real world coordinates to lidar
#
# p - area using real world coordinate
# l - area using lidar coordinate(pointcloud)
# max_limit - maximum area limit of a lidar scann in real world coordinates
def position_relative_to_pointcloud(l1, l2, p1, p2, center_lim, max_lim):
    center_l = tuple((a+b)/2 for a, b in zip(l1,l2))
    return [
        ((p1[0] - center_lim[0]) * ((l1[0] - l2[0]) / (max_lim[2][0] - max_lim[0][0])) + center_l[0],
            (p1[1] - center_lim[1]) * ((l1[1] - l2[1]) / (max_lim[2][1] - max_lim[0][1])) + center_l[1]),
        ((p2[0] - center_lim[0]) * ((l1[0] - l2[0]) / (max_lim[2][0] - max_lim[0][0])) + center_l[0],
            (p2[1] - center_lim[1]) * ((l1[1] - l2[1]) / (max_lim[2][1] - max_lim[0][1])) + center_l[1]),
    ]
#print(position_relative_to_pointcloud((-20,20), (-10,30), (1,3), (2,2), (2.5,2.5), [(5,5),(0,5),(0,0),(5,0)]))
#print(position_relative_to_pointcloud((-3299999, 4608018), (-3200001, 4687153), (61.47620866851029, 8.961138281887507), (61.95241733702057, 6.8508373935926645), (61, 11), [(61.95241733702057, 15.125349549679886), (60.04758266297943, 15.125349549679886), (60.04758266297943, 6.8746504503201145), (61.95241733702057, 6.8746504503201145)]))

# check if lidar points is within range of the area selected
def inArea(position, areaRange):
    x, y, _ = position # position to be checked
    #print((areaRange[0][0])," < ",x," < ",(areaRange[1][0])," and ",(areaRange[0][1])," < ",(y)," < ",(areaRange[1][1])," ",((areaRange[0][0]) < x < (areaRange[1][0])) and ((areaRange[0][1]) < (y) < (areaRange[1][1])))
    if ((areaRange[0][0]) < x < (areaRange[1][0])) and ((areaRange[0][1]) < y < (areaRange[1][1])):
        return True
    else:
        return False

# find distance between two points
def distance(point1, point2):
    y1, x1 = point1
    y2, x2 = point2
    return math.sqrt(abs(y2 - y1)**2 + abs(x2 - x1)**2)

# find the closest point in json list
def closest_points(point, list, coords_taken):
    closest_point_found = None
    closest_dist = float('inf')
    for current_point in list:
        dist = distance(point, current_point['properties']['sub_div_center'])
        if dist < closest_dist and current_point not in coords_taken:
            closest_dist = dist
            closest_point_found = current_point
    return closest_point_found

# Calculation of height in an area from las files
def find_height(points):
    if len(points) == 0:
        return [0]
    height_differences = [] # final container for height data

    # sort the points
    sorted_coords = sorted(points, key=lambda coord: (coord[0], coord[1]))

    # group the sorted points that has the same xy- coordinates together
    groupCoords = [list(group) for key, group in groupby(sorted_coords, key=lambda coord: (coord[0], coord[1]))]

    # loop through the groups to find the difference in height
    for group in groupCoords:
        if len(group) < 2 or group[0] == group[1]:
            continue # jump over iteration if there is only one coordinate or lidar registered the same point twice

        # find max and min height
        min_height = min(coords[2] for coords in group)
        max_height = max(coords[2] for coords in group)

        # difference between them
        difference = max_height - min_height
        height_differences.append(difference)

    # list of thickness in an area
    return height_differences

def calculate_corners(lat, lng, area_offset):
    """Calculate corners of polygon based on a center coordinate

    Arguments:
    lat -- center latitude
    lng -- center longitude
    """
    # From https://stackoverflow.com/questions/7477003/calculating-new-longitude-latitude-from-old-n-meters
    # Formulas:
    offset_lng, offset_lat = area_offset
    lat_pos = lat + (offset_lat * METER)
    lng_pos = lng + (offset_lng * METER) / cos(lat * (pi / 180))

    lat_neg = lat - (offset_lat * METER)
    lng_neg = lng - (offset_lng * METER) / cos(lat * (pi / 180))

    return [
        (lat_neg, lng_pos), # top left
        (lat_pos, lng_pos), # top right
        (lat_pos, lng_neg), # bottom right
        (lat_neg, lng_neg)  # bottom left
    ]

# separate the zones into smaller area, into a grid
def define_gridareas(lat, lng, area_offset, grid_size):
    # container for an area turned into a grid of areas
    grided_area = []

    # find the main area's corner positions
    main_area = calculate_corners(lat, lng, area_offset)

    offset_lng, offset_lat = area_offset

    # find the main area's range size
    area_size = (main_area[2][0] - main_area[0][0], main_area[0][1] - main_area[2][1])
    # find each subareas vector to it's center
    dist_to_subcenter = (area_size[0]/(grid_size*2), area_size[1]/(grid_size*2))

    # find subareas size relative to main area
    subarea_offset_lng = offset_lng/grid_size
    subarea_offset_lat = offset_lat/grid_size

    # find the center coordinates of each area in grid to find the corner areas
    for y in (range(grid_size)):
        relative_size_lat = y / grid_size
        for x in (range(grid_size)):
            relative_size_lng = x / grid_size
            lat_pos = main_area[3][0] + relative_size_lat * area_size[0] + dist_to_subcenter[0]
            lng_pos = main_area[3][1] + relative_size_lng * area_size[1] + dist_to_subcenter[1]

            # use the center of sub areas to find the corner of each subarea
            subarea_offset = (subarea_offset_lng, subarea_offset_lat)
            corners = calculate_corners(lat_pos, lng_pos, subarea_offset)
            grided_area.append(corners)

    return grided_area

# separate the zones into smaller area, into a grid
def define_grid_lidardata(max_area, grid_size, points):
    # container for an area turned into a grid of areas
    grided_area = []

    bottom_left, top_right = max_area

    # find subareas size relative to the full size of data
    subarea_offset_lng = (top_right[0]-bottom_left[0])/grid_size
    subarea_offset_lat = (top_right[1]-bottom_left[1])/grid_size

    # find the center coordinates of each area in grid to find the corner areas
    for y in (range(grid_size)):
        relative_size_lat = y / grid_size
        for x in (range(grid_size)):
            relative_size_lng = x / grid_size
            bottom_left_range = (subarea_offset_lng * (relative_size_lng - 1 / 2) + bottom_left[0],
                                 subarea_offset_lat * (relative_size_lat - 1 / 2) + bottom_left[1])

            top_right_range = (subarea_offset_lng * relative_size_lng + bottom_left[0],
                               subarea_offset_lat * relative_size_lat + bottom_left[1])

            area_range = [bottom_left_range, top_right_range]

            area = list(filter(lambda point_position: inArea(point_position, area_range), points))

            height_in_area = find_height(area)

            grided_area.append(height_in_area)

    return grided_area
#[1,2,2,3,4,5,6,3,4,6,8,9,5,3,5.7,8,5,3]
#print(define_gridareas(60,10,1500,4))
#print(define_gridareas(3435693.5,6299200.0, 400000,4))